from jsonmathpy.interfaces.iterator import IIterator
from jsonmathpy.interfaces.lexer import ILexer
from jsonmathpy.interfaces.node_provider import INodeProvider
from jsonmathpy.interfaces.parser import IParser
from jsonmathpy.interfaces.tokens import ITokenProvider
from jsonmathpy.models.node_keys import ConfigurationModels

class ParserBuilder:

    def __init__(self, 
                    lexer:              ILexer, 
                    parser:             IParser, 
                    token_provider:     ITokenProvider, 
                    node_provider:      INodeProvider, 
                    iterator:           IIterator, 
                    node_keys:          ConfigurationModels
                ):
        self.token_provider     = token_provider
        self.node_provider      = node_provider
        self.node_keys          = node_keys
        self.lexer              = lexer
        self.parser             = parser
        self.iterator           = iterator

    def build(self, text) -> IParser:
        # Instantiate providers:
        token_provider_instance = self._build_token_provider()
        node_provider_instance  = self._build_node_provider()

        # Configure node provider to the custom, user defined, variables and functions methods.
        node_provider_instance.set_matcher(self.node_keys if self.node_keys != None else [])

        # Wrap the text we wish to parse around an iterator object:
        character_iter_instance = self._build_iterator(text)

        # Build Lexer with user input:
        lexer = self._build_lexer(token_provider_instance, character_iter_instance)

        # Tokenize the input:
        tokens = lexer.tokenize()

        # Wrap the tokens around the iterable once more:
        tokens_iter_instance = self._build_iterator(tokens)

        # Build the Parser object from tokens generated by user input:
        parser = self.parser(node_provider_instance, tokens_iter_instance)

        # Return an instance of Parser object, which implements IParser:
        return parser

    def _build_node_provider(self) -> INodeProvider:
        return self.node_provider()

    def _build_token_provider(self) -> ITokenProvider:
        return self.token_provider()

    def _build_lexer(self, token_provider: ITokenProvider, iterator: IIterator) -> ILexer:
        return self.lexer(token_provider, iterator)

    def _build_parser(self, token_provider: ITokenProvider, iterator: IIterator) -> ILexer:
        return self.parser(token_provider, iterator)
    
    def _build_iterator(self, type) -> IIterator:
        return self.iterator(type)